--This script loads a large set of data and then adds data to show how the process works.

USE [Demo_AnalyzeAndTuneStatistics];
GO

SET LANGUAGE us_english;

IF SCHEMA_ID(N'Sales') IS NULL
	EXEC('CREATE SCHEMA Sales AUTHORIZATION [dbo];');
GO

DROP TABLE IF EXISTS Sales.OrderHeader;
GO

CREATE TABLE Sales.OrderHeader
(
	Id					INT IDENTITY(1,1)
	,SubmittedDate		DATE NOT NULL
	,CustomerId			INT NOT NULL
	,DeliveryAddressId	INT NOT NULL
	,BillingAddressId	INT NOT NULL
	,StatusCode			CHAR(3) NOT NULL
	,Originator			VARCHAR(128) NOT NULL
	,Notes				NVARCHAR(1024) NOT NULL
	CONSTRAINT PK_OrderHeader PRIMARY KEY CLUSTERED
	(
		Id
	)
);
GO

CREATE OR ALTER PROC dbo.sp_BasicChecks
AS
BEGIN
	EXEC sp_spaceused N'Sales.OrderHeader';

	SELECT	COUNT(DISTINCT SubmittedDate)
	FROM	Sales.OrderHeader;

	;WITH CTE_01 AS
	(
		SELECT		SubmittedDate
					,COUNT(*) AS [Count]
		FROM		Sales.OrderHeader
		GROUP BY	SubmittedDate
	)
	SELECT	MIN([Count]) AS MinCount
			,MAX([Count]) AS MaxCount
	FROM	CTE_01;

	SELECT		so.[name] AS [table_name]
				,st.[name] AS [stats_name]
				,st.has_persisted_sample
				,st.no_recompute
				,stp.last_updated
				,stp.modification_counter
				,stp.persisted_sample_percent
				,stp.[rows]
				,stp.rows_sampled
				,stp.steps
				,COALESCE(ROUND((CAST(rows_sampled AS real)/CAST([rows] AS real)*100),2),0) AS last_sampling_rate
	FROM		sys.objects so
	INNER JOIN	sys.stats st
	ON			so.[object_id] = st.[object_id]
	CROSS APPLY sys.dm_db_stats_properties(st.[object_id],st.stats_id) stp
	WHERE		OBJECT_SCHEMA_NAME(so.[object_id]) = N'Sales';
END;
GO

/*
Case 1
*/

TRUNCATE TABLE Sales.OrderHeader;

DECLARE @v_SubmittedDate				date = '20220701'
		,@v_NbOrderHeadersDayCurrent	int
		,@v_NbOrderHeadersDayTarget		int;

WHILE @v_SubmittedDate<='20221231'
BEGIN
	SET @v_NbOrderHeadersDayTarget=ROUND(RAND()*100000,0);

	/*Note: linear sales volume (between 40'000 and 60'000 sales header per day)*/
	IF @v_NbOrderHeadersDayTarget<40000
		SET @v_NbOrderHeadersDayTarget=50000;

	IF @v_NbOrderHeadersDayTarget>60000
		SET @v_NbOrderHeadersDayTarget=50000;

	INSERT INTO Sales.OrderHeader
	(
		SubmittedDate
		,CustomerId
		,DeliveryAddressId
		,BillingAddressId
		,StatusCode
		,Originator
		,Notes
	)
	SELECT	TOP(@v_NbOrderHeadersDayTarget)
			@v_SubmittedDate
			,ROUND(RAND()*1111,0)
			,ROUND(RAND()*2222,0)
			,ROUND(RAND()*3333,0)
			,'val'
			,'website'
			,N'Great, one more sale!'
	FROM	sys.messages;

	SET @v_SubmittedDate=DATEADD(DAY,1,@v_SubmittedDate);
END;

/*Note: manually update stats as auto updates are disabled*/
UPDATE STATISTICS Sales.OrderHeader(PK_OrderHeader);

/*Note: associated stats will be built with fullscan*/
CREATE INDEX IX_Date ON Sales.OrderHeader(SubmittedDate);
/*Note: fallback to default sampling rate for the demo*/
UPDATE STATISTICS Sales.OrderHeader(IX_Date);
GO

EXEC dbo.sp_BasicChecks;

EXEC [dbo].[sp_AnalyzeAndTuneStatistics]
	@p_SchemaName=N'Sales'
	,@p_ObjectName=N'OrderHeader'
	,@p_StatName=N'IX_Date'
	,@p_ShowDetails=1;

/*
Case 2
*/

DROP INDEX IX_Date ON Sales.OrderHeader;
TRUNCATE TABLE Sales.OrderHeader;

DECLARE @v_SubmittedDate				date = '20220701'
		,@v_NbOrderHeadersDayCurrent	int
		,@v_NbOrderHeadersDayTarget		int;

WHILE @v_SubmittedDate<='20221231'
BEGIN
	SET @v_NbOrderHeadersDayTarget=ROUND(RAND()*100000,0);

	INSERT INTO Sales.OrderHeader
	(
		SubmittedDate
		,CustomerId
		,DeliveryAddressId
		,BillingAddressId
		,StatusCode
		,Originator
		,Notes
	)
	SELECT	TOP(@v_NbOrderHeadersDayTarget)
			@v_SubmittedDate
			,ROUND(RAND()*1111,0)
			,ROUND(RAND()*2222,0)
			,ROUND(RAND()*3333,0)
			,'val'
			,'website'
			,N'Great, one more sale!'
	FROM	sys.messages;

	SET @v_SubmittedDate=DATEADD(DAY,1,@v_SubmittedDate);
END;

/*Note: manually update stats as auto updates are disabled*/
UPDATE STATISTICS Sales.OrderHeader(PK_OrderHeader);

/*Note: associated stats will be built with fullscan*/
CREATE INDEX IX_Date ON Sales.OrderHeader(SubmittedDate);
/*Note: fallback to default sampling rate for the demo*/
UPDATE STATISTICS Sales.OrderHeader(IX_Date);
GO

EXEC dbo.sp_BasicChecks;

EXEC [dbo].[sp_AnalyzeAndTuneStatistics]
	@p_SchemaName=N'Sales'
	,@p_ObjectName=N'OrderHeader'
	,@p_StatName=N'IX_Date'
	,@p_ShowDetails=1;

SELECT		TOP(8000)
			*
FROM		Sales.OrderHeader
WHERE		SubmittedDate = '20221006'
ORDER BY	CustomerId;

EXEC [dbo].[sp_AnalyzeAndTuneStatistics]
	@p_SchemaName=N'Sales'
	,@p_ObjectName=N'OrderHeader'
	,@p_StatName=N'IX_Date'
	,@p_TryTuning=1
	,@p_ShowDetails=1;

/*
Case 3
*/

DROP INDEX IX_Date ON Sales.OrderHeader;
TRUNCATE TABLE Sales.OrderHeader;

DECLARE @v_SubmittedDate				date = '20180101'
		,@v_NbOrderHeadersDayCurrent	int
		,@v_NbOrderHeadersDayTarget		int;

WHILE @v_SubmittedDate<='20221231'
BEGIN
	SET @v_NbOrderHeadersDayTarget=ROUND(RAND()*10000,0);

	/*Note: linear sales volume (between 4'000 and 6'000 sales header per day)*/
	IF @v_NbOrderHeadersDayTarget<4000
		SET @v_NbOrderHeadersDayTarget=5000;

	IF @v_NbOrderHeadersDayTarget>6000
		SET @v_NbOrderHeadersDayTarget=5000;

	INSERT INTO Sales.OrderHeader
	(
		SubmittedDate
		,CustomerId
		,DeliveryAddressId
		,BillingAddressId
		,StatusCode
		,Originator
		,Notes
	)
	SELECT	TOP(@v_NbOrderHeadersDayTarget)
			@v_SubmittedDate
			,ROUND(RAND()*1111,0)
			,ROUND(RAND()*2222,0)
			,ROUND(RAND()*3333,0)
			,'val'
			,'website'
			,N'Great, one more sale!'
	FROM	sys.messages;

	SET @v_SubmittedDate=DATEADD(DAY,1,@v_SubmittedDate);
END;

/*Note: manually update stats as auto updates are disabled*/
UPDATE STATISTICS Sales.OrderHeader(PK_OrderHeader);

/*Note: associated stats will be built with fullscan*/
CREATE INDEX IX_Date ON Sales.OrderHeader(SubmittedDate);
/*Note: fallback to default sampling rate for the demo*/
UPDATE STATISTICS Sales.OrderHeader(IX_Date);
GO

EXEC dbo.sp_BasicChecks;

EXEC [dbo].[sp_AnalyzeAndTuneStatistics]
	@p_SchemaName=N'Sales'
	,@p_ObjectName=N'OrderHeader'
	,@p_StatName=N'IX_Date'
	,@p_TryTuning=1
	,@p_ShowDetails=1;

/*
Case 4
*/

DROP INDEX IX_Date ON Sales.OrderHeader;
TRUNCATE TABLE Sales.OrderHeader;

DECLARE @v_SubmittedDate				date = '20180101'
		,@v_NbOrderHeadersDayCurrent	int
		,@v_NbOrderHeadersDayTarget		int;

WHILE @v_SubmittedDate<='20221231'
BEGIN
	SET @v_NbOrderHeadersDayTarget=ROUND(RAND()*100000,0);

	INSERT INTO Sales.OrderHeader
	(
		SubmittedDate
		,CustomerId
		,DeliveryAddressId
		,BillingAddressId
		,StatusCode
		,Originator
		,Notes
	)
	SELECT	TOP(@v_NbOrderHeadersDayTarget)
			@v_SubmittedDate
			,ROUND(RAND()*1111,0)
			,ROUND(RAND()*2222,0)
			,ROUND(RAND()*3333,0)
			,'val'
			,'website'
			,N'Great, one more sale!'
	FROM	sys.messages;

	SET @v_SubmittedDate=DATEADD(DAY,1,@v_SubmittedDate);
END;

/*Note: manually update stats as auto updates are disabled*/
UPDATE STATISTICS Sales.OrderHeader(PK_OrderHeader);

/*Note: associated stats will be built with fullscan*/
CREATE INDEX IX_Date ON Sales.OrderHeader(SubmittedDate);
/*Note: fallback to default sampling rate for the demo*/
UPDATE STATISTICS Sales.OrderHeader(IX_Date);
GO

EXEC dbo.sp_BasicChecks;

EXEC [dbo].[sp_AnalyzeAndTuneStatistics]
	@p_SchemaName=N'Sales'
	,@p_ObjectName=N'OrderHeader'
	,@p_StatName=N'IX_Date'
	,@p_TryTuning=1
	,@p_ShowDetails=1;
